"""String filters for Baygon.
Each filter is a class that implements a filter method.
A filter is used to modify `stdout` and `stderr` before they are tested.
"""

import inspect
import re
import sys
from abc import ABC, abstractmethod
from collections.abc import Sequence
from functools import lru_cache

from tinykernel import TinyKernel

from .error import InvalidFilterError


class Filter(ABC):
    """Base class for filters."""

    @abstractmethod
    def apply(self, value: str) -> str:
        """Apply the filter to a value."""
        return value

    def filter(self, value: str) -> str:
        """Apply the filter to a value."""
        return self.apply(value)

    def __init__(self, *args, **kwargs):
        """Initialize the filter.
        Keyword arguments:

        input: Boolean Is the filter applied to the input?
        """
        self.input = kwargs.get("input", False)

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def __call__(self, value: str) -> str:
        return self.filter(value)

    @classmethod
    def name(cls):
        """Return the name of the filter."""
        return cls.__name__.split("Filter", maxsplit=1)[1].lower()


class FilterNone(Filter):
    """Filter that does nothing."""

    def apply(self, value: str) -> str:
        return value


class FilterUppercase(Filter):
    """Filter for uppercase strings.
    >>> f = FilterUppercase()
    >>> f('hello')
    'HELLO'
    """

    def apply(self, value: str) -> str:
        return value.upper()


class FilterLowercase(Filter):
    """Filter for lowercase strings.
    >>> f = FilterLowercase()
    >>> f('HELLO')
    'hello'
    """

    def apply(self, value: str) -> str:
        return value.lower()


class FilterTrim(Filter):
    """Filter for trimmed strings.
    >>> f = FilterTrim()
    >>> f(' hello   ')
    'hello'
    """

    __output__ = True

    def apply(self, value: str) -> str:
        return value.strip()


class FilterIgnoreSpaces(Filter):
    """Filter for strings with no spaces.
    >>> f = FilterIgnoreSpaces()
    >>> f('hello   world')
    'helloworld'
    """

    def apply(self, value: str) -> str:
        return value.replace(" ", "")


class FilterReplace(Filter):
    """Filter for strings with simple replacements.
    >>> f = FilterReplace('hello', 'world')
    >>> f('hello world')
    'world world'
    """

    def __init__(self, pattern: str, replacement: str):
        super().__init__()
        self.pattern = pattern
        self.replacement = replacement

    def apply(self, value: str) -> str:
        return value.replace(self.pattern, self.replacement)


class FilterRegex(Filter):
    """Filter for strings using regular expressions.
    >>> f = FilterRegex('[aeiou]', '-')
    >>> f('hello world')
    'h-ll- w-rld'
    """

    def __init__(self, pattern: str, replacement: str):
        super().__init__()
        self.pattern = pattern
        self.replacement = replacement
        self.regex = re.compile(pattern)

    def apply(self, value: str) -> str:
        return self.regex.sub(self.replacement, value)


class FilterEval(Filter):
    """Filter for evaluating mustaches in strings."""

    def __init__(self, start: str = "{{", end: str = "}}", init: list = None):
        super().__init__()
        self._mustache = re.compile(f"{start}(.*?){end}")
        self._kernel = TinyKernel()

        init += [
            "from math import *",
            "from random import *",
            "from statistics import *",
            "from baygon.eval import iter",
        ]

        for item in init:
            self._kernel(item)

    def apply(self, value: str) -> str:
        """Evaluate mustaches in a string."""
        pos = 0
        ret = ""
        for match in self._mustache.finditer(value):
            ret += value[pos : match.start()]
            ret += str(self.exec(match.group(1)))
            pos = match.end()
        ret += value[pos:]
        return ret

    def exec(self, code: str):
        """Execute code in the kernel."""

        # Inject context to custom functions
        code = re.sub(r"((?<=\b)iter\(.*?)(\))", f"\\1,ctx={hash(code)}\\2", code)

        # Workaround to get the value of assignments
        try:
            self._kernel("_ = " + code)
            return self._kernel.glb["_"]
        except SyntaxError:
            return self._kernel(code)

    def __repr__(self):
        return f"{self.__class__.__name__}({self._mustache.pattern})"


class Filters(Filter, Sequence):
    """A sequence of filters."""

    def __init__(self, filters=None):
        super().__init__()
        self._filters = self._parse_filter(filters)

    def _parse_filter(self, filters):

        if filters is None:
            return []
        if isinstance(filters, Filter):
            return [filters]
        if isinstance(filters, Filters):
            return list(filters._filters)
        if isinstance(filters, dict):
            instances = []
            for name, args in filters.items():
                if not isinstance(args, list):
                    args = [args]
                instances.append(FilterFactory(name, *args))
            return instances

        raise InvalidFilterError(f"Invalid type for filters: {type(filters)}")

    def __getitem__(self, index):
        return self._filters[index]

    def __len__(self):
        return len(self._filters)

    def extend(self, filters):
        """Extend the filters with another Filters object."""
        self._filters.extend(self._parse_filter(filters))
        return self

    def apply(self, value: str) -> str:
        for filter_ in self._filters:
            value = filter_.filter(value)
        return value

    def __repr__(self):
        return f"{self.__class__.__name__}<{self._filters}>"


class FilterFactory:
    """Factory for filters."""

    @classmethod
    @lru_cache()
    def filters(cls):
        """Helper to get all filters by their name."""
        fmap = {}
        for _, member in inspect.getmembers(sys.modules[__name__]):
            if not inspect.isclass(member) or not hasattr(member, "name"):
                continue
            if member.name() == "base" or len(member.name()) < 2:
                continue
            fmap[member.name()] = member
        return fmap

    def __new__(cls, name, *args, **kwargs) -> Filter:
        if name not in cls.filters():
            raise ValueError(f"Unknown matcher: {name}")
        return cls.filters()[name](*args, **kwargs)



"""YAML Schema for Baygon test files."""

from voluptuous import (
    All,
    Any,
    Boolean,
    Coerce,
    ExactSequence,
    Exclusive,
    Optional,
    Required,
    Self,
)
from voluptuous import Schema as VSchema
from voluptuous.humanize import validate_with_humanized_errors

from .id import TrackId

Value = Any(str, All(Any(int, float, All(bool, Coerce(int))), Coerce(str)))


class ToList:
    """Convert the given value to a list."""

    def __call__(self, v):
        return [v]


class ToDict:
    """Convert the given value to a dict."""

    def __init__(self, key):
        self.key = key

    def __call__(self, v):
        return {self.key: v}


class Join:
    """Join the given values."""

    def __init__(self, sep):
        self.sep = sep

    def __call__(self, v):
        return self.sep.join(v)


# Eval module
evaluate = {
    Optional("eval", description="Eval mustaches"): Any(
        {
            Optional("start", default="{{"): str,
            Optional("end", default="}}"): str,
            Optional("init", str): Any(All(str, ToList()), [str]),
        },
        All(Any(Boolean(), None), lambda x: {"init": []}),
    )
}

# Global test filters
filters = {
    Optional("uppercase"): Boolean(),
    Optional("lowercase"): Boolean(),
    Optional("trim"): Boolean(),
    Optional(Any("ignorespaces", "ignore-spaces")): Boolean(),
    Optional("regex"): ExactSequence([str, str]),
    Optional("replace"): ExactSequence([str, str]),
}

# Test case
case = VSchema(
    {
        Optional("filters", default={}): filters,
        Any("equals", "regex", "contains"): Value,
        Optional("not"): [{Required(Any("equals", "regex", "contains")): Value}],
        Optional("expected", description="Expected value when used with regex"): Value,
    }
)

# Nested test cases
match = Any(
    All(Value, ToDict("equals"), ToList()),
    All(case, ToList()),
    [All(Value, lambda x: ToDict("equals")), case],
)


common = {
    Optional("name", default="", description="Test Name"): str,
    Optional("executable", default=None, description="Path to the executable"): Any(
        str, None
    ),
    Optional(
        Exclusive(
            "points", "points_or_weight", description="Points given for this test"
        )
    ): Any(float, int),
    Optional(
        Exclusive("weight", "points_or_weight", description="Weight of the test")
    ): Any(float, int),
    Optional("min-points", default=0.1): Any(float, int),
}

test = VSchema(
    {
        Optional("args", default=[]): [Value],
        Optional("env", default={}): {str: Value},
        Optional("stdin", default=""): Any(Value, None),
        Optional("stdout", default=[]): match,
        Optional("stderr", default=[]): match,
        Optional("repeat", default=1): int,
        Optional("exit"): Any(int, str, bool),
    }
).extend(common)

Num = TrackId()

group = VSchema(
    {Required("tests"): All(Num.down(), [All(Any(Self, test), Num.next())], Num.up())}
).extend(common)

cli = {
    Optional("verbose"): Any(int),
    Optional("report"): str,
    Optional("format"): Any("json", "yaml"),
    Optional("table", default=False): Boolean(),
}


def Schema(data, humanize=False):  # noqa: N802
    """Validate the given data against the schema."""
    schema = (
        VSchema(
            {
                Optional("name"): str,
                Optional("version", default=2): Any(1, 2),
                Optional("filters", default={}): filters,
                Required("tests"): All(
                    Num.reset(), [All(Any(test, group), Num.next())]
                ),
                Optional("points"): Any(float, int),
                Optional("min-points", default=0.1): Any(float, int),
            }
        )
        .extend(common)
        .extend(evaluate)
        .extend(cli)
    )
    if humanize:
        return validate_with_humanized_errors(data, schema)
    return schema(data)



"""Test suite."""

import json
from pathlib import Path

import yaml

from .error import ConfigError, InvalidExecutableError
from .executable import Executable
from .filters import FilterEval, FilterNone, Filters
from .id import Id
from .matchers import InvalidExitStatus, MatcherFactory
from .schema import Schema
from .score import compute_points


def find_testfile(path=None):
    """Recursively find the tests description file."""
    if not path:
        path = Path(".")
    elif isinstance(path, str):
        path = Path(path)

    path = path.resolve(strict=True)

    if path.is_file():
        return path

    for filename in ["baygon", "t", "test", "tests"]:
        for ext in ["json", "yml", "yaml"]:
            f = path.joinpath(f"{filename}.{ext}")
            if f.exists():
                return f

    # Recursively search in parent directories
    if path.parent == path:  # Test if root directory
        return None

    return find_testfile(path.parent)


def load_config(path=None):
    """Load a configuration file (can be YAML or JSON)."""
    path = find_testfile(path)

    if not path.exists():
        raise ConfigError(f"Couldn't find and configuration file in '{path.resolve()}'")

    with open(path, "rt", encoding="utf-8") as fp:
        if path.suffix in [".yml", ".yaml"]:
            return Schema(yaml.safe_load(fp))
        if path.suffix in [".json"]:
            return Schema(json.load(fp))

    raise ConfigError(f"Unknown file extension '{path.suffix}' for '{path}'")


class BaseMixin:
    """Base mixin to prevent super() failure.
    Ensure it will be the last MRO (Method Resolution Order)."""

    def __init__(self, *args, **kwargs):
        self.parent = kwargs.get("parent", None)


class FilterMixin(BaseMixin):
    """Mixin for tests that can have filters."""

    def __init__(self, *args, **kwargs):
        config = args[0]

        # Configure filters
        if "parent" in kwargs and hasattr(kwargs["parent"], "filters"):
            self.filters = kwargs["parent"].filters
        else:
            self.filters = Filters()

        if "filters" in config:
            self.filters.extend(config["filters"])

        # Eval
        if "parent" in kwargs and hasattr(kwargs["parent"], "eval"):
            self.eval = kwargs["parent"].eval
        else:
            self.eval = FilterNone()

        if isinstance(config.get("eval"), dict):
            self.eval = FilterEval(**config["eval"])

        super().__init__(*args, **kwargs)


class ExecutableMixin(BaseMixin):
    """Mixin for tests that have an executable."""

    def __init__(self, *args, **kwargs):
        config = args[0]
        self.cwd = kwargs["parent"].cwd

        if hasattr(kwargs["parent"], "executable"):
            executable = kwargs["parent"].executable
        else:
            executable = None

        if executable is not None and config["executable"] is not None:
            raise InvalidExecutableError("Executable can't be overridden")

        if config["executable"] is not None:
            exe = self.cwd.joinpath(config["executable"]).resolve(strict=True)
            self.executable = Executable(exe)
        else:
            self.executable = Executable(executable)

        super().__init__(*args, **kwargs)


class NamedMixin(BaseMixin):
    """Mixin for tests having a name."""

    def __init__(self, *args, **kwargs):
        config = args[0]

        self.name = config["name"]
        self.id = Id(config["test_id"])
        self.points = config.get("points", 1)

        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}<{self.id}. {self.name}>"


class GroupMixin(BaseMixin):
    """Mixin for tests that can have sub-tests."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        config = args[0]
        self.tests = []
        for test in config["tests"]:
            if "tests" in test:
                self.tests.append(TestGroup(test, parent=self))
            else:
                self.tests.append(TestCase(test, parent=self))

    def __len__(self):
        return len(self.tests)

    def __getitem__(self, item):
        return self.tests[item]

    def run(self, flatten=False):
        """Run the tests."""
        if flatten:
            issues = []
            for test in self.tests:
                run = test.run()
                issues += run if isinstance(run, list) else [run]
        else:
            issues = [test.run() for test in self.tests]
        return issues

    def get_points(self):
        """Return the total number of points."""
        return sum(test.points for test in self.tests)


class TestCase(NamedMixin, ExecutableMixin, FilterMixin):
    """A single test."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        config = args[0]

        self.args = config.get("args", [])
        self.env = config.get("env", {})

        self.stdin = config.get("stdin", "")
        self.stdout = config.get("stdout", [])
        self.stderr = config.get("stderr", [])

        self.exit = config.get("exit", None)

        self.repeat = config.get("repeat", 1)

        self.config = config
        self.output = None
        self.issues = []

        self.filtered_args = None
        self.filtered_exit = None
        self.filtered_stdin = None

    def _eval(self, data):
        if not self.eval:
            return data
        if isinstance(data, list):
            return [self.eval(arg) for arg in data]
        return self.eval(data)

    def run(self, hook=None):
        """Run the tests."""
        if not isinstance(self.executable, Executable):
            raise InvalidExecutableError(
                f"Test {self.id}, not a valid executable: {self.executable}"
            )

        self.issues = []
        for _ in range(self.repeat):
            self.filtered_args = self._eval(self.args)
            self.filtered_stdin = self._eval(self.stdin)
            self.filtered_exit = (
                int(self._eval(str(self.exit))) if self.exit is not None else None
            )

            self.output = output = self.executable.run(
                *self.filtered_args, stdin=self.filtered_stdin, hook=hook
            )

            for on in ["stdout", "stderr"]:
                for case in getattr(self, on):
                    self._match(on, case, output)
                    if "not" in case:  # Check for negative match
                        self._match(on, case["not"], output, inverse=True)

            if (
                self.filtered_exit is not None
                and self.filtered_exit != output.exit_status
            ):
                self.issues.append(
                    InvalidExitStatus(
                        self.filtered_exit, output.exit_status, on="exit", test=self
                    )
                )

        return self.issues

    def _match(self, on, case, output, inverse=False):
        """Match the output."""
        if "filters" in case:
            filters = Filters(self.filters).extend(case["filters"])
        else:
            filters = self.filters

        for key in MatcherFactory.matchers():
            if key in case:
                out = filters(getattr(output, on))
                matcher = MatcherFactory(key, self._eval(case[key]), inverse=inverse)
                issue = matcher(out, on=on, test=self)
                if issue:
                    self.issues.append(issue)


class TestGroup(NamedMixin, ExecutableMixin, FilterMixin, GroupMixin):
    """A group of tests."""


class TestSuite(ExecutableMixin, FilterMixin, GroupMixin):
    """Test suite."""

    __test__ = False  # Don't run this class as a test

    def __init__(self, data: dict = None, path=None, executable=None, cwd=None):
        if isinstance(data, dict):
            self.config = Schema(data)
            cwd = Path.cwd()
        else:
            self.path = find_testfile(path)
            cwd = self.path
            self.config = load_config(self.path)

        compute_points(self.config)

        self.name = self.config.get("name", "Test Suite")
        self.version = self.config.get("version")

        class Root:
            def __init__(self, executable):
                self.executable = Executable(executable)
                self.cwd = cwd.resolve(strict=True).parent

        super().__init__(self.config, parent=Root(executable))



"""Errors for Baygon"""


class BaygonError(Exception):
    """Base class for Baygon errors"""


class ConfigError(BaygonError):
    """Raised when a config value is not valid"""


class InvalidExecutableError(BaygonError):
    """Raised when an executable is not found"""


class InvalidFilterError(BaygonError):
    """Raised when a filter is not found"""



import re
import shlex


def escape_argument(arg):
    """Escape a command line argument.

    >>> print(escape_argument("hello"))
    hello
    >>> print(escape_argument("hello world"))
    'hello world'
    >>> print(escape_argument("hello'world'"))
    'hello'"'"'world'"'"''
    """
    return shlex.quote(arg)


def create_command_line(args):
    """Create a command line from a list of arguments.

    >>> print(create_command_line(["echo", "hello world"]))
    echo 'hello world'
    """
    escaped_args = [escape_argument(str(arg)) for arg in args]
    return " ".join(escaped_args)


class GreppableString(str):
    """A string that can be parsed with regular expressions."""

    def grep(self, pattern: str, *args) -> bool:
        """Return True if the pattern is found in the string.

        >>> GreppableString("hello world").grep("w.{3}d")
        ['world']
        >>> GreppableString("hello world").grep(r"\b[a-z]{4}\b")
        []
        """
        return re.findall(pattern, self, *args)

    def contains(self, value: str) -> bool:
        """Return True if the value is found in the string.

        >>> GreppableString("hello world").contains("world")
        True
        >>> GreppableString("hello world").contains("earth")
        False
        """
        return value in self



import inspect
import re
import sys
from abc import ABC, abstractmethod
from functools import lru_cache


class InvalidCondition:
    """Invalid test case condition."""

    def __init__(self, value, expected, on=None, test=None, **kwargs):
        self.on = on
        self.test = test
        self.value = value
        self.expected = expected

    def __str__(self):
        return (
            f'Expected value "{self.expected}" on {self.on}, '
            f'but got "{self.value}" instead.'
        )

    def __repr__(self):
        return f"{self.__class__.__name__}<{str(self)}>"


class InvalidExitStatus(InvalidCondition):
    """Invalid exit status error."""

    def __str__(self):
        return f"Invalid exit status: " f"{self.value} != {self.expected}."


class InvalidContains(InvalidCondition):
    """Invalid contains error."""

    def __str__(self):
        return (
            f"Output {self.on} does not contain {self.expected}. "
            f'Found "{self.value}" instead.'
        )


class InvalidRegex(InvalidCondition):
    """Invalid regex error."""

    def __str__(self):
        return (
            f"Output '{self.on}' does not match /{self.expected}/ "
            f'on "{self.value}".'
        )


class InvalidEquals(InvalidCondition):
    """Invalid equals error."""

    def __str__(self):
        value = "(empty)" if not self.value else f"'{self.value}'"
        return f"Output {value} does not equal '{self.expected}' on " f"{self.on}."


class MatchBase(ABC):
    """Base class for all matchers."""

    def __init__(self, inverse=False, **kwargs):
        """Initialize the matcher."""
        self.inverse = inverse

    @classmethod
    def name(cls):
        """Return the name of the filter."""
        return cls.__name__.split("Match", maxsplit=1)[1].lower()

    @abstractmethod
    def __call__(self, value, **kwargs):
        """Match the value against the condition."""
        raise NotImplementedError


class MatchRegex(MatchBase):
    """Match a regex."""

    def __init__(self, pattern, **kwargs):
        """Initialize the matcher."""
        self.pattern = re.compile(pattern)
        super().__init__(**kwargs)

    def __call__(self, value, **kwargs):
        if (not self.pattern.findall(value)) ^ self.inverse:
            return InvalidRegex(value, self.pattern.pattern, **kwargs)
        return None


class MatchContains(MatchBase):
    """Match if a string contains a specific value."""

    def __init__(self, contains, **kwargs):
        """Initialize the matcher."""
        self.contains = contains
        super().__init__(**kwargs)

    def __call__(self, value, **kwargs):
        if (self.contains not in value) ^ self.inverse:
            return InvalidContains(value, self.contains, **kwargs)
        return None


class MatchEquals(MatchBase):
    """Match if a string contains a specific value."""

    def __init__(self, equal, **kwargs):
        """Initialize the matcher."""
        self.equal = equal
        super().__init__(**kwargs)

    def __call__(self, value, **kwargs):
        if (self.equal != value) ^ self.inverse:
            return InvalidEquals(value, self.equal, **kwargs)
        return None


class MatcherFactory:
    """Factory for matchers."""

    @classmethod
    @lru_cache()
    def matchers(cls):
        """Helper to get all matchers by their name."""
        fmap = {}
        for _, member in inspect.getmembers(sys.modules[__name__]):
            if not inspect.isclass(member) or not hasattr(member, "name"):
                continue
            if member.name() == "base":
                continue
            fmap[member.name()] = member
        return fmap

    def __new__(cls, name, *args, **kwargs) -> MatchBase:
        if name not in cls.matchers():
            raise ValueError(f"Unknown matcher: {name}")
        return cls.matchers()[name](*args, **kwargs)



"""Main CLI command."""

import logging
import os
import sys
import time

import click
from rich.box import SQUARE_DOUBLE_HEAD
from rich.console import Console
from rich.rule import Rule
from rich.style import Style
from rich.table import Table

from . import TestCase, TestGroup, TestSuite, __copyright__, __version__
from .error import InvalidExecutableError
from .helpers import create_command_line

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("baygon")

console = Console()


def display_pad(pad=0):
    """Display a pad of spaces to align nested output."""
    click.secho("." * pad if pad > 0 else "  ", nl=False, dim=pad == 0)


def test_name_length(test):
    """Compute the length of test name."""
    return len(f"{test.id.pad()}Test {test.id}: {test.name}")


def display_test_name(test):
    """Display the name of a test."""
    click.secho(f"{test.id.pad()}Test {test.id}: ", nl=False, bold=True)
    click.secho(f"{test.name}", nl=False, bold=False)


def display_table(data):
    grey_line_style = Style(color="grey37")
    title_style = Style(bold=False, color="white")
    table = Table(
        title="Test Summary",
        title_style=title_style,
        border_style=grey_line_style,
        box=SQUARE_DOUBLE_HEAD,
    )

    table.add_column("ID", justify="left")
    table.add_column("Test Name", justify="left", style=grey_line_style)
    table.add_column("Points", justify="center", style=grey_line_style)
    table.add_column("Status", justify="center", style=grey_line_style)

    for test in data:
        status = ""
        points = ""
        if test.get("status") is not None:
            status = "[green]PASS[/green]" if test["status"] else "[red]FAIL[/red]"

            points = f"{test['earned']}/{test['points']}"

        table.add_row(
            str(test["id"]), (" " * len(test["id"])) + test["name"], points, status
        )
    console.print(table)


class OneLineExceptionFormatter(logging.Formatter):
    """A formatter that displays the exception on a single line."""

    def format_exception(self, exc_info):
        """Format an exception."""
        result = super().format_exception(exc_info)
        return repr(result)

    def format(self, record):
        """Format a log record."""
        result = super().format(record)
        if record.exc_text:
            result = result.replace("\n", "")
        return result


class Runner:
    """Test runner."""

    def __init__(self, executable=None, config=None, **kwargs):
        self.executable = executable
        self.limit = kwargs.get("limit", -1)
        click.secho(f"Using configuration file: {config}")
        self.test_suite = TestSuite(path=config, executable=self.executable)

        # Forward cli arguments to the test suite
        if "verbose" in kwargs and kwargs["verbose"] != 0:
            self.test_suite.config["verbose"] = kwargs["verbose"]
        if "report" in kwargs and kwargs["report"] is not None:
            self.test_suite.config["report"] = kwargs["report"]
        if "format" in kwargs and kwargs["format"] is not None:
            self.test_suite.config["format"] = kwargs["format"]
        if "table" in kwargs and kwargs["table"] is not None:
            self.test_suite.config["table"] = kwargs["table"]

        self.verbose = kwargs.get("verbose", 0)
        if self.verbose > 0:
            click.echo("Verbose level set to " + str(self.verbose))

        if self.verbose > 3:
            self._init_logger("DEBUG")

        self.align_column = 0

        self.failures = 0
        self.successes = 0
        self.skipped = 0
        self.run_time = 0

        self.points_earned = 0
        self.points_total = 0

        self.summary = []

    def _init_logger(self, loglevel):
        handler = logging.StreamHandler()
        formatter = OneLineExceptionFormatter(logging.BASIC_FORMAT)
        handler.setFormatter(formatter)
        root = logging.getLogger()
        root.setLevel(os.environ.get("LOGLEVEL", loglevel))
        root.addHandler(handler)

    def get_report(self):
        self.test_suite.get_points()
        return {
            "failures": self.failures,
            "successes": self.successes,
            "skipped": self.skipped,
            "total": self.failures + self.successes + self.skipped,
            "time": self.run_time,
            "points": {
                "total": self.points_total,
                "earned": self.points_earned,
            },
        }

    def run(self):
        """Run the tests."""
        start_time = time.time()
        self.align_column = self._max_length(self.test_suite) + 10

        self.failures = 0
        self.successes = 0
        self.skipped = 0
        self.points_total = 0
        self.points_earned = 0

        self._traverse_group(self.test_suite)

        if self.test_suite.config.get("table"):
            display_table(self.summary)

        tests = self.failures + self.successes + self.skipped
        seconds = round(time.time() - start_time, 2)
        self.run_time = seconds
        console.print(f"\nRan [bold]{tests}[/bold] tests in [bold]{seconds} s[/bold].")

        if self.points_total > 0:
            click.secho(
                f"Points: {self.points_earned}/{self.points_total}",
                bold=True,
            )

        if self.failures > 0:
            ratio = 100 - round(
                self.failures / (self.failures + self.successes) * 100, 2
            )
            click.secho(
                (f"{self.failures} failed, {self.successes} " f"passed ({ratio}% ok)."),
                fg="yellow",
                bold=True,
            )
            click.secho("\nfail.", fg="red", bold=True)
        else:
            click.secho("\nok.", fg="green")

        if self.skipped > 0:
            click.secho(
                (
                    f"{self.skipped} test(s) skipped, "
                    "some executables may be missing."
                ),
                fg="yellow",
            )

        report = self.test_suite.config.get("report")
        report_format = self.test_suite.config.get("format")
        if report:
            if not report_format:
                if report.endswith(".yaml"):
                    report_format = "yaml"
                else:
                    report_format = "json"
            save_report(self.get_report(), report, report_format)

        return self.failures

    def _max_length(self, tests):
        length = 0
        for test in tests:
            length = max(test_name_length(test), length)
            if isinstance(test, TestGroup):
                length = max(self._max_length(test), length)
        return length

    def hook_ran_command(self, cmd, stdin, stdout, stderr, exit_status, **kwargs):
        """Capture executed command for logging purpose."""
        # If element have a space, wrap it with quotes

        cmdline = create_command_line(cmd)
        cin = f' <<< "{stdin.decode("utf-8")}"' if stdin else ""
        self.ran_commands.append(
            f"{cmdline}{cin} (exited with: {exit_status})\n{stdout}{stderr}"
        )

    def pad(self, text, length):
        """Shift text to right
        >>> pad("hello", 5)"
        '     hello'
        >>> pad("hello\nworld", 3)"
        '   hello\n   world'
        """
        return "\n".join(" " * length + line for line in text.splitlines())

    def display_test_verbose(self, test, issues, verbose=0):  # noqa: C901

        p = test
        name = []
        while True:
            if p is None or getattr(p, "name", None) is None:
                break
            if p.name:
                name.append(p.name)
            p = p.parent
        name = " / ".join(reversed(name))

        if verbose >= 2:
            console.print("\n")
            console.print(Rule(f"Test {test.id}: {name}", align="left", style="bold"))

        if verbose >= 3:
            click.secho(
                self.pad("".join([f"$ {cmd}\n" for cmd in self.ran_commands]), 2),
                dim=True,
            )

        if test.status == "skipped":
            if verbose > 0:
                click.secho(" SKIPPED", fg="yellow")
            else:
                click.secho("S", fg="yellow", nl=False)
            self.skipped += 1
            return

        if test.status == "passed":
            if verbose > 0:
                click.secho(" PASSED", fg="green")
            else:
                click.secho(".", fg="green", nl=False)
            self.successes += 1
            return

        if test.status == "failed":
            if verbose > 0:
                click.secho(" FAILED", fg="red")
            else:
                click.secho("F", fg="red", nl=False)
            self.failures += 1

        for k, issue in enumerate(issues):
            click.secho(str(issue))

    def _run_test(self, test):
        self.ran_commands = []
        hook = self.hook_ran_command
        issues = test.run(hook)
        if issues is None:
            test.status = "skipped"
        elif not issues:
            test.status = "passed"
        else:
            test.status = "failed"

        self.display_test_verbose(test, issues, self.verbose)

        earned_points = test.points if test.status == "passed" else 0
        self.points_total += test.points
        self.points_earned += earned_points
        self.summary.append(
            {
                "id": test.id,
                "name": test.name,
                "status": test.status,
                "points": test.points,
                "earned": earned_points,
            }
        )

    def _traverse_group(self, tests):
        for test in tests:
            if self.limit > 0 and self.failures > self.limit:
                break

            if isinstance(test, TestGroup):
                self.summary.append(
                    {
                        "id": test.id,
                        "name": test.name,
                    }
                )
                self._traverse_group(test)
            elif isinstance(test, TestCase):
                self._run_test(test)
        return self.failures


def save_report(data, filename, format):
    """Save the report to a file."""
    if format == "json":
        import json

        with open(filename, "w") as fp:
            json.dump(data, fp, indent=2, sort_keys=True)
    elif format == "yaml":
        import yaml

        with open(filename, "w") as fp:
            yaml.dump(data, fp)


def version():
    """Display the version."""
    print(f"Baygon version {__version__} {__copyright__}")


@click.command()
@click.argument("executable", required=False, type=click.Path(exists=True))
@click.option("--version", is_flag=True, help="Shows version")
@click.option("-v", "--verbose", count=True, help="Shows more details")
@click.option("-l", "--limit", type=int, default=-1, help="Limit errors to N")
@click.option("-d", "--debug", is_flag=True, default=False, help="Debug mode")
@click.option("-r", "--report", type=click.Path(), help="Report file")
@click.option("-t", "--table", is_flag=True, default=False, help="Summary table")
@click.option(
    "-f", "--format", type=click.Choice(["json", "yaml"]), help="Report format"
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True),
    help="Choose config file (.yml or .json)",
)
def cli(debug, **kwargs):
    """Baygon functional test runner."""
    if kwargs.get("version"):
        version()
        sys.exit(0)

    if debug:
        logger.setLevel(logging.DEBUG)
        logger.debug("Debug mode enabled.")
        runner = Runner(**kwargs)
        runner.run()
    else:
        try:
            runner = Runner(**kwargs)
            runner.run()
        except InvalidExecutableError as error:
            click.secho(f"\nError: {error}", fg="red", bold=True, err=True)
            sys.exit(1)

    click.echo("")


if __name__ == "__main__":
    cli()



"""Baygon is a tool to run tests on executables."""

from .executable import Executable
from .schema import Schema
from .suite import TestCase, TestGroup, TestSuite

__all__ = ["TestCase", "TestGroup", "TestSuite", "Executable", "Schema"]

try:
    from . import version

    __version__ = version.version
except ImportError:
    __version__ = None

__year__ = 2022
__copyright__ = f"Copyright {__year__} HEIG-VD"



""" Module used to compute the score of a test case.
  Used in academic.
"""

from decimal import ROUND_DOWN, ROUND_HALF_UP, Decimal, getcontext


def float_or_int(value):
    """Return a float or an integer.
    >>> float_or_int(1.0)
    1
    >>> float_or_int(1.1)
    1.1
    """
    if value == int(value):
        return int(value)
    return float(value)


def distribute(values, total, min_value):
    """Distrubute the values given using a minimum step to ensure
    the total sum is respected.

    >>> distribute([1, 2, 3, 4], 10, 0.001)
    [1, 2, 3, 4]
    >>> distribute([1, 1, 1, 1], 100, 0.01)
    [25, 25, 25, 25]
    >>> distribute([12.5, 19.7, 42.1, 8.9], 100, 0.2)
    [15, 23.7, 50.6, 10.7]
    >>> distribute([100, 100, 200, 200], 50, 1)
    [8, 8, 17, 17]
    """
    getcontext().prec = 28
    total_weight = sum(values)
    total = Decimal(str(total))
    min_value = Decimal(str(min_value))
    decimal_values = [Decimal(str(v)) for v in values]
    allocations = [v / Decimal(str(total_weight)) * total for v in decimal_values]

    # Round down to the nearest multiple of min_value
    quantizer = min_value
    allocations_rounded = [
        a.quantize(quantizer, rounding=ROUND_HALF_UP) for a in allocations
    ]

    # Adjust the allocations to match the total
    total_allocated = sum(allocations_rounded)
    difference = total - total_allocated

    # If the difference is not zero, adjust the allocations
    if difference != Decimal("0"):
        # Compute the number of units to adjust
        units_to_adjust = int(
            (difference / min_value).to_integral_value(rounding=ROUND_HALF_UP)
        )
        # Sort the allocations by their remainders
        remainders = [
            a - a.quantize(quantizer, rounding=ROUND_DOWN) for a in allocations
        ]
        if units_to_adjust > 0:
            indices = sorted(
                range(len(values)), key=lambda i: remainders[i], reverse=True
            )
            adjustment = min_value
        else:
            indices = sorted(range(len(values)), key=lambda i: remainders[i])
            adjustment = -min_value
            units_to_adjust = abs(units_to_adjust)

        for i in range(units_to_adjust):
            idx = indices[i % len(indices)]
            allocations_rounded[idx] += adjustment

    return [float_or_int(a) for a in allocations_rounded]


def assign_points(test, parent=None):  # noqa: C901
    """Assign points recursively to each test in the structure."""
    min_point = test.get("min-points", parent.get("min-points", 1) if parent else 1)

    # Case 6: Default points if there are weights or points somewhere
    default_point = 1 if has_weights_or_points(test) else 0

    # Set default weight to 10 if not specified
    if "weight" not in test and "points" not in test and "tests" in test:
        test["weight"] = 10

    # If 'points' is not in test, compute it based on parent's points and weights
    if "points" not in test:
        if "weight" in test and parent and "points" in parent:
            total_siblings_weight = parent.get("_total_weights", 0)
            if total_siblings_weight == 0:
                total_siblings_weight = sum(
                    t.get("weight", 10) for t in parent.get("tests", [])
                )
                parent["_total_weights"] = total_siblings_weight
            test["points"] = test["weight"] / total_siblings_weight * parent["points"]
            test["points"] = float_or_int(
                Decimal(str(test["points"])).quantize(
                    Decimal(str(min_point)), rounding=ROUND_HALF_UP
                )
            )
        else:
            test["points"] = default_point

    # If there are subtests, assign points to them
    if "tests" in test:
        fixed_points = sum(
            subtest.get("points", 0) for subtest in test["tests"] if "points" in subtest
        )
        weights = []
        subtests_to_distribute = []
        for subtest in test["tests"]:
            if "points" in subtest:
                continue  # Skip subtests that already have points
            elif "weight" in subtest:
                weights.append(subtest["weight"])
                subtests_to_distribute.append(subtest)
            else:
                # Case 2 and 6: Default weight is 10 if not specified
                subtest["weight"] = 10
                weights.append(subtest["weight"])
                subtests_to_distribute.append(subtest)

        points_to_distribute = test["points"] - fixed_points

        if weights and points_to_distribute > 0:
            allocated_points = distribute(weights, points_to_distribute, min_point)
            for subtest, points in zip(subtests_to_distribute, allocated_points):
                subtest["points"] = points
        elif not weights and points_to_distribute > 0:
            # Case 3: No weights but points are given to all tests
            equally_divided_point = points_to_distribute / len(test["tests"])
            for subtest in test["tests"]:
                if "points" not in subtest:
                    subtest["points"] = float_or_int(
                        Decimal(str(equally_divided_point)).quantize(
                            Decimal(str(min_point)), rounding=ROUND_HALF_UP
                        )
                    )

    # Recursive call for subtests
    if "tests" in test:
        for subtest in test["tests"]:
            assign_points(subtest, parent=test)

    # Clean up temporary keys
    if "_total_weights" in test:
        del test["_total_weights"]


def has_weights_or_points(test):
    """Check if there are weights or points in the test or its subtests.
    >>> has_weights_or_points({"weight": 10})
    True
    """
    if "weight" in test or "points" in test:
        return True
    if "tests" in test:
        return any(has_weights_or_points(subtest) for subtest in test["tests"])
    return False


def compute_points(data):
    """Compute points for the entire structure."""
    # Case 4: If no weights or points exist anywhere, do nothing
    if not has_weights_or_points(data):
        data["compute-score"] = False
        return data

    # Case 5: Set compute-score to true
    data["compute-score"] = True

    # Case 7: If total points are given at root but no other info, set default weight
    if "points" in data and "tests" in data:
        for test in data["tests"]:
            if "weight" not in test and "points" not in test:
                test["weight"] = 10

    assign_points(data)
    return data



"""Helper functions used in `eval` filter.
These functions are injected into the kernel and made available from
the mustaches templates.

Examples:

    >>> reset()
    >>> iter()
    0
    >>> iter()
    1
    >>> iter()
    2
    >>> iter(100, 10)
    100
    >>> iter(100, 10)
    110
    >>> iter(ctx='foo')
    0
    >>> iter(ctx='foo')
    1
    >>> reset()
    >>> iter()
    0
    >>> iter(100, 10)
    100
"""

_context = {}


def reset():
    """Reset the context."""
    _context.clear()


def iter(start=0, step=1, ctx=None):
    """Custom iterator for eval input filter."""
    ctx = (start, step, ctx)
    _context[ctx] = _context.get(ctx, start - step) + step
    return _context[ctx]
